#include <iostream>
#include <fstream>
#include <sys/stat.h>

#include <arc/IString.h>
#include <arc/StringConv.h>
#include <arc/Logger.h>
#include <arc/IniConfig.h>
#include <arc/ArcRegex.h>
#include <arc/credential/Credential.h>
#include <arc/credential/VOMSUtil.h>
#include <arc/credential/VOMSAttribute.h>

#define AC_POLICY_PARAM_NAME "ac_policy"
#define ARC_LOGGER

static void usage(char *pname) {
	std::cerr << Arc::IString("Usage: %s [-N] -P <user proxy> -L <A-REX local> [-c <configfile>]",pname) << std::endl;
}

int main(int argc, char *argv[]) {
	int opt;
	int no_ac_success = 0;
	int verbose = 0;
	const char *user_proxy_f = NULL;
	const char *job_local_f = NULL;
	const char *config_f = NULL;

	// log
	static Arc::Logger logger(Arc::Logger::rootLogger, "arc-vomsac-check");

	// Parse options 
	while ((opt = getopt(argc, argv, "NP:L:c:v")) != -1) {
		switch (opt) {
			case 'N':
				no_ac_success = 1;
				break;
			case 'P':
				user_proxy_f = optarg;
				break;
			case 'L':
				job_local_f = optarg;
				break;
			case 'c':
				config_f = optarg;
				break;
			case 'v':
				verbose = 1;
				break;
			default:
				usage(argv[0]);
				return EXIT_FAILURE;
		}
	}

	// check required
	if ( !user_proxy_f ) {
#ifdef ARC_LOGGER
		logger.msg(Arc::ERROR,"User proxy file is required but not specified");
#else
		std::cerr << "User proxy file is required but not specified" << std::endl;
#endif
		usage(argv[0]);
		return EXIT_FAILURE;
	}
	if ( !job_local_f ) {
#ifdef ARC_LOGGER
		logger.msg(Arc::ERROR,"A-REX local parameters file is required");
#else
		std::cerr << "A-REX local parameters file is required" << std::endl;
#endif
		usage(argv[0]);
		return EXIT_FAILURE;
	}
	if ( !config_f ) {
		config_f = "/etc/arc.conf";
	}

	// read information about the job used by the A-REX
	// and determine selected queue
	std::string queue;
 	std::ifstream job_local;
	job_local.open(job_local_f, std::ios::in);
	if ( job_local.is_open() ) {
		std::string line;
		while ( ! job_local.eof() ){
			getline(job_local,line);
			if ( ! line.compare(0,6,"queue=") ) {
				queue = line.substr(6);
				break;
#ifdef ARC_LOGGER
				logger.msg(Arc::INFO,"making decision using queue: %s",queue);
#else
				if ( verbose ) std::cerr << Arc::IString("INFO: making decision using queue: %s",queue) << std::endl;
#endif
			}
		}
	} else {
#ifdef ARC_LOGGER
		logger.msg(Arc::ERROR,"can not read information from A-REX local parameters file");
#else
		std::cerr << "Can not read information from A-REX local parameters file" << std::endl;
#endif
		return EXIT_FAILURE;
	}
	job_local.close();

	// Parse INI configuraion file and get queue access_policy parameters
	Arc::IniConfig cfg(config_f);
	if ( ! cfg ) {
#ifdef ARC_LOGGER
		logger.msg(Arc::ERROR,"Can not parse config file");
#else
		std::cerr << "Can not parse config file" << std::endl;
#endif
		return EXIT_FAILURE;
	}
	
	std::string qqueue = '"' + queue + '"';
	Arc::XMLNode qparams = cfg["queue/" + queue];
	if ( ! (bool)qparams ) {
		for(Arc::XMLNode qnode = cfg["queue"];(bool)qnode;++qnode) {
			if ( (std::string)qnode["id"] == qqueue || (std::string)qnode["name"] == qqueue ) {
				qparams = qnode;
				break;
			}
		}
	}
	if ( ! (bool)qparams) {
#ifdef ARC_LOGGER
		logger.msg(Arc::ERROR,"Can not find queue '%s' in configuration file",queue);
#else
		std::cerr << Arc::IString("Can not find queue '%s' in configuration file",queue) << std::endl;
#endif
		return EXIT_FAILURE;
	}

	std::vector<Arc::RegularExpression> access_policies;
	for ( Arc::XMLNode pnode = qparams[AC_POLICY_PARAM_NAME];(bool)pnode;++pnode) {
		std::string acp = (std::string)pnode;
		if ( acp.compare(0,5,"VOMS:") ) {
			// TODO: more precise to regular expressions
			// TODO: implement an NOT MATCH regular expression
			std::string regex = ".*" + Arc::trim(acp.substr(6)," /\"") + "[^/]*";
			access_policies.push_back(Arc::RegularExpression(regex));
		}
	}
	if ( access_policies.empty() ) {
#ifdef ARC_LOGGER
		logger.msg(Arc::INFO,"No access policy to check, returning success");
#else
		if ( verbose ) std::cerr << "INFO:No access policy to check, returning success" << std::endl;
#endif
		return EXIT_SUCCESS;
	}

	// CA Cert directory requires to work with proxy
	std::string ca_dir;
	ca_dir = (std::string)cfg["common"]["x509_cert_dir"];
	if (ca_dir.empty()) {
		ca_dir = "/etc/grid-security/certificates";
	} else {
		ca_dir = Arc::trim(ca_dir,"\"");
	}
	struct stat statFInfo;
	if ( stat(ca_dir.c_str(),&statFInfo) ) {
#ifdef ARC_LOGGER
		logger.msg(Arc::ERROR,"CA DIR %s does nor exists", ca_dir);
#else
		std::cerr << Arc::IString("CA DIR %s does nor exists", ca_dir) << std::endl;
#endif
		return EXIT_FAILURE;
	}

	// construct ARC credentials object
	Arc::Credential holder(user_proxy_f, "", ca_dir, "");
	if (! holder.GetVerification()) {
#ifdef ARC_LOGGER
		logger.msg(Arc::ERROR,"User proxy certificate is not valid");
#else
		std::cerr << "User proxy certificate is not valid" << std::endl;
#endif
		return EXIT_FAILURE;
	}

	// get VOMS AC from proxy certificate
#ifdef ARC_LOGGER
		logger.msg(Arc::VERBOSE,"Getting VOMS AC for: %s", holder.GetDN());
#else
		if ( verbose) std::cerr << Arc::IString("VERBOSE: Getting VOMS AC for: %s", holder.GetDN()) << std::endl;
#endif
	std::vector<Arc::VOMSACInfo> voms_attributes;
	const Arc::VOMSTrustList vomscert_trust_dn;

	if ( ! Arc::parseVOMSAC(holder, "", "", vomscert_trust_dn, voms_attributes, false) ) {
#ifdef ARC_LOGGER
		logger.msg(Arc::ERROR,"Error parsing VOMS AC");
#else
		std::cerr << "Error parsing VOMS AC" << std::endl;
#endif
		if ( no_ac_success ) return EXIT_SUCCESS;
		return EXIT_FAILURE;
	}

	// look for FQAN match one of the access policies 
	for (std::vector<Arc::VOMSACInfo>::iterator iAC = voms_attributes.begin(); iAC != voms_attributes.end(); iAC++) {
#ifdef ARC_LOGGER
		logger.msg(Arc::VERBOSE,"VOMS extension for VO: %s",iAC->voname);
#else
		if ( verbose ) std::cerr << Arc::IString("VOMS extension for VO: %s",iAC->voname) << std::endl;
#endif
		for (int acnt = 0; acnt < iAC->attributes.size(); acnt++ ) {
			std::string fqan = iAC->attributes[acnt];
#ifdef ARC_LOGGER
			logger.msg(Arc::VERBOSE,"Found attribute: %s",fqan);
#else
			if ( verbose ) std::cerr << Arc::IString("Found attribute: %s",fqan) << std::endl;
#endif
			for (std::vector<Arc::RegularExpression>::iterator iP = access_policies.begin(); iP != access_policies.end(); iP++) {
				// match policy regex 
				if ( iP->match(fqan) ) {
#ifdef ARC_LOGGER
					logger.msg(Arc::VERBOSE,"FQAN '%s' IS match '%s'",fqan,iP->getPattern());
#else
					if (verbose ) std::cerr << Arc::IString("FQAN '%s' match '%s'",fqan,iP->getPattern()) << std::endl;
#endif
					return EXIT_SUCCESS;
				} else {
#ifdef ARC_LOGGER
					logger.msg(Arc::VERBOSE,"FQAN '%s' DO NOT match '%s'",fqan,iP->getPattern());
#else
					if (verbose) std::cerr << Arc::IString("FQAN '%s' DO NOT match '%s'",fqan,iP->getPattern()) << std::endl;
#endif
				}
			}
		}
	}
#ifdef ARC_LOGGER
	logger.msg(Arc::ERROR,"Queue access policy do not allow provided FQANs");
#else
	std::cerr << "Queue access policy do not allow provided FQANs" << std::endl;
#endif
	return EXIT_FAILURE;
}

